#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 28 16:10:26 2021

@author: Nathan Cross
"""

from datetime import datetime, date
from numpy import (asarray, float64, int64)
from os import listdir, path, walk
from pandas import DataFrame, read_csv
from wonambi.attr import Annotations
from ..utils.logs import create_logger, create_logger_outfile
                        
def export_sleepstats(xml_dir, out_dir, subs = 'all', sessions = 'all', 
               rater = None,  times = None, arousal_name = ['Arousal', 'Arou'],
               logger = create_logger('Export macro stats')):
    
    ### 0.a Set up logging
    flag = 0
    logger.info('')
    
    ## 1. Get lights on & lights off file
    if isinstance(times, str):
        times = read_csv(times, sep='\t')
    elif isinstance(times, DataFrame):
        None
    else:
        logger.critical('To export macro stats, lights off and lights on times are requried.')
        logger.info('Check documentation for how to export macro statistics:')
        logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
        logger.info('-' * 10)
    
    # 2.a Get subjects
    if isinstance(subs, list):
        None
    elif subs == 'all':
        try:
            subs = next(walk(xml_dir))[1]
        except:
            logger.critical(f"{xml_dir} doesn't exist!") 
            return
        if len(subs) == 0:
            logger.critical(f'{xml_dir} is empty!') 
            return
    else:
        logger.info('')
        logger.critical("'subs' must either be an array of Participant IDs or = 'all' ")
        return
    subs.sort()
    
    # 2.b. Get sessions
    sub_ses = {}
    if isinstance(sessions, list):
        for sub in subs:
            sub_ses[sub] = sessions
    elif sessions == 'all':
        for sub in subs:
            try:
                session = next(walk(f'{xml_dir}/{sub}'))[1]
            except Exception as e:
                logger.critical(f"{xml_dir}/{sub} either doesn't exist or is empty.")
                return
    
            session.sort()
            sub_ses[sub] = session
    else:
        logger.info('')
        logger.critical("'sessions' must either be a list of Session IDs or = 'all' ")
        return
    
    for s, sub in enumerate(sub_ses): #for each participant...
        if not sub_ses[sub]:
            logger.warning(f'No visits found in {xml_dir}/{sub}. Skipping..')
            flag +=1
            continue
        for v, ses in enumerate(sub_ses[sub]): #for each visit...
            logger.debug(f'Export macro statistics for {sub}, {ses}')

            # 3.d. Get proper annotations file
            xml_file = [x for x in listdir(xml_dir +  '/' + sub + '/' + ses ) 
                        if x.endswith('.xml') if not x.startswith('.')]    
            if len(xml_file) == 0:                
                logger.warning(f"No sleep staging file found for {sub}, {ses}. "
                               "Skipping..")
                flag+=1
                continue
            elif len(xml_file) > 1:
                logger.warning(f"> 1 sleep staging file was found for {sub}, "
                               f"visit {ses} - to select the correct file you "
                               "must define the variable 'keyword'. Skipping..")
                flag+=1
                continue
            else:
                xml_file_path = f'{xml_dir}/{sub}/{ses}/{xml_file[0]}'
           
                # Search for sub and ses in tracking sheet
                l_times = times[times['sub']==sub]
                if l_times.size == 0:
                    logger.warning(f"Participant not found in column 'sub' in "
                                   f"'tracking.tsv' for {sub}, {ses}. This is "
                                   "required for LightsOFF and LightsON times. "
                                   "Skipping...")
                    flag +=1
                    continue
                l_times = l_times[l_times['ses']==ses]
                if l_times.size == 0:
                    logger.warning(f"Session not found in column 'ses' in "
                                   f"'tracking.tsv' for {sub}, {ses}. This is "
                                   "required for LightsOFF and LightsON times. "
                                   "Skipping...")
                    flag +=1
                    continue
                
                ## Lights OFF
                lights_off = l_times['loff']
                lights_off = asarray(lights_off.dropna())
                if lights_off.size == 0:
                    logger.warning(f"Lights Off time not found in 'tracking.tsv' "
                                   f"for {sub}, {ses}. Skipping...")
                    flag +=1
                    continue
                else:
                    if isinstance(lights_off[0],int64):
                        lights_off = float(lights_off[0])
                    else:
                        try:
                            lights_off = lights_off.astype(float64)[0]
                        except:
                            logger.warning("Error reading Lights Off time in "
                                          f"'tracking.tsv' for {sub}, {ses}. "
                                           "Skipping...")
                            flag +=1
                            continue
                
                ## Lights ON
                lights_on = l_times['lon']
                lights_on = asarray(lights_on.dropna())
                if lights_on.size == 0:
                    logger.warning(f"Lights On time not found in 'tracking.tsv' "
                                   f"for {sub}, {ses}. Skipping...")
                    flag +=1
                    continue
                else:
                    if isinstance(lights_on[0],int64):
                        lights_on = float(lights_on[0])
                    else:
                        try:
                            lights_on = lights_on.astype(float64)[0]
                        except:
                            logger.warning("Error reading Lights On time in "
                                          f"'tracking.tsv' for {sub}, {ses}. "
                                           "Skipping...")
                            flag +=1
                            continue
                
                
                # Export sleep stats
                annot = Annotations(xml_file_path, rater_name=rater)
                annot.export_sleep_stats(f'{out_dir}/{sub}/{ses}/{sub}_{ses}_sleepstats.csv', 
                                         lights_off, lights_on)
                
                # Export Arousal density (if exists)
                xml_file = [file for file in listdir(f'{xml_dir}/{sub}/{ses}/') if 
                             '.xml' in file][0]
                annot = Annotations(f'{xml_dir}/{sub}/{ses}/{xml_file}')

                try:
                    file = f'{xml_dir}/{sub}/{ses}/{sub}_{ses}_arousals.csv' 
                    annot.export_events(file, evt_type = arousal_name,
                                        stage = ['NREM1', 'NREM2', 'NREM3', 'REM']) 
                except:
                    logger.warning(f'Error exporting arousals for {sub}, {ses}.')
                    flag += 1
 
    ### 3. Check completion status and print
    if flag == 0:
        logger.debug('Macro statistics export finished without ERROR.')  
    else:
        logger.warning(f'Macro statistics export finished with {flag}+ WARNINGS. See log for details.')
    
    return   


def sleepstats_from_csvs(xml_dir, out_dir, subs = 'all', sessions = 'all', 
                         arousal_name = ['Arousal', 'Arou'],
                         logger = create_logger('Export macro stats')):
    
    ### 0.a Set up logging
    flag = 0
    logger.info('')    
    
    # 1. Get subjects
    if isinstance(subs, list):
        None
    elif subs == 'all':
            subs = next(walk(xml_dir))[1]
    else:
        logger.info('')
        logger.critical("'subs' must either be an array of Participant IDs or = 'all' ")
        return
    subs.sort()
    
    # 2. Get sessions
    if isinstance(sessions, list):
        None
    elif sessions == 'all':
        sessions = []
        for i, sub in enumerate(subs):
            sessions.append(next(walk(f'{xml_dir}/{sub}'))[1])
        sessions = sorted(set([x for y in sessions for x in y]))
    else:
        logger.info('')
        logger.critical("'sessions' must either be an array of Session IDs or = 'all' ")
        return
        
   
    
    # 3. Run sleep statistic parameter extraction
    # header = ['ses', 'TIB_min', 'TotalWake_min', 'SL_min', 'WASOintra_min', 
    #           'Wmor_min', 'TSP_min', 'TST_min', 'SE_%', 'N1_min', 'N2_min', 
    #           'N3_min', 'REM_min', 'W_%tsp', 'N1_%tsp', 'N2_%tsp', 'N3_%tsp', 
    #           'REM_%tsp', 'SSI', 'SFI', 'SL_toN2_min', 'SL_toN3_min', 'SL_toREM_min', 
    #           'SL_toNREM_5m_min', 'SL_toNREM_10m_min', 'SL_toN3_5m_min', 'SL_toN3_10m_min']

    for v, ses in enumerate(sessions): #for each visit...
        df = DataFrame(data=None, index=subs)
        
        for sub in subs: #for each participant...

            logger.debug(f'Extracting macro stats from {sub}, {ses}')  
            
            data_file = f'{xml_dir}/{sub}/{ses}/{sub}_{ses}_sleepstats.csv'
            
            if not path.exists(data_file):
                logger.warning(f"No (exported) staging data found for {sub}, {ses}. ")
                flag += 1
                continue
            else:
            
                data = read_csv(data_file, sep=',', delimiter=None, 
                            header=1, index_col=0) 
                df.loc[sub, f'TIB_min_{ses}'] =             round(float(data['Value 2']['Total dark time (Time in bed)']),3)
                df.loc[sub, f'TotalWake_min_{ses}'] =       round(float(data['Value 2']['Wake duration']),3)
                df.loc[sub, f'SL_min_{ses}'] =              round(float(data['Value 2']['Sleep latency']),3)
                df.loc[sub, f'WASOintra_min_{ses}'] =       round(float(data['Value 2']['Wake after sleep onset']),3)
                df.loc[sub, f'Wmor_min_{ses}'] =            round(float(data['Value 2']['Wake, morning']),3)
                df.loc[sub, f'TSP_min_{ses}'] =             round(float(data['Value 2']['Total sleep period']),3)
                df.loc[sub, f'TST_min_{ses}'] =             round(float(data['Value 2']['Total sleep time']),3)
                df.loc[sub, f'SE_%tsp_{ses}'] =             round(float(data['Value 1']['Sleep efficiency']),3)
                df.loc[sub, f'NREM1_min_{ses}'] =              round(float(data['Value 2']['N1 duration']),3)
                df.loc[sub, f'NREM2_min_{ses}'] =              round(float(data['Value 2']['N2 duration']),3)
                df.loc[sub, f'NREM3_min_{ses}'] =              round(float(data['Value 2']['N3 duration']),3)
                df.loc[sub, f'REM_min_{ses}'] =             round(float(data['Value 2']['REM duration']),3)
                df.loc[sub, f'W_%tsp_{ses}'] =              round(float(data['Value 1']['W % TSP']),3)
                df.loc[sub, f'NREM1_%tsp_{ses}'] =             round(float(data['Value 1']['N1 % TSP']),3)
                df.loc[sub, f'NREM2_%tsp_{ses}'] =             round(float(data['Value 1']['N2 % TSP']),3)
                df.loc[sub, f'NREM3_%tsp_{ses}'] =             round(float(data['Value 1']['N3 % TSP']),3)
                df.loc[sub, f'REM_%tsp_{ses}'] =            round(float(data['Value 1']['REM % TSP']),3)
                df.loc[sub, f'SSI_{ses}'] =                 round(float(data['Value 1']['Switch index H']),3)
                df.loc[sub, f'SFI_{ses}'] =                 round(float(data['Value 1']['Sleep fragmentation index H']),3)
                df.loc[sub, f'SL_toN2_min_{ses}'] =         round(float(data['Value 2']['Sleep latency to N2']),3)
                df.loc[sub, f'SL_toN3_min_{ses}'] =         round(float(data['Value 2']['Sleep latency to N3']),3)
                df.loc[sub, f'SL_toREM_min_{ses}'] =        round(float(data['Value 2']['Sleep latency to REM']),3)
                df.loc[sub, f'SL_toNREM_5m_min_{ses}'] =    round(float(data['Value 2']['Sleep latency to consolidated NREM, 5 min']),3)
                df.loc[sub, f'SL_toNREM_10m_min_{ses}'] =   round(float(data['Value 2']['Sleep latency to consolidated NREM, 10 min']),3)
                df.loc[sub, f'SL_toN3_5m_min_{ses}'] =      round(float(data['Value 2']['Sleep latency to consolidated N3, 5 min']),3)
                df.loc[sub, f'SL_toN3_10m_min_{ses}'] =     round(float(data['Value 2']['Sleep latency to consolidated N3, 10 min']),3)
                
                
                ## add Arousal density
                arou_file = f'{xml_dir}/{sub}/{ses}/{sub}_{ses}_arousals.csv' 
                if not path.exists(arou_file):
                    logger.warning(f"No (exported) arousal data found for {sub}, {ses}. ")
                    flag += 1
                    continue
                else:
                    #TST
                    arou = read_csv(arou_file, sep = ',', skiprows=[0], delimiter=None)
                    df.loc[sub, f'Arousal_index_TST_{ses}'] = round(float(arou.shape[0]/(df.loc[sub, f'TST_min_{ses}'])*60),3)                                        
                    #REM
                    arou_rem = arou[arou['Stage']=="REM"]
                    df.loc[sub, f'Arousal_index_REM_{ses}'] = round(float(arou_rem.shape[0]/(df.loc[sub, f'REM_min_{ses}'])*60),3)
                    #NREM
                    arou_nrem = arou[arou['Stage'].str.contains("NREM", na=False)]
                    NREM_dur = df.loc[sub, f'NREM1_min_{ses}'] + df.loc[sub, f'NREM2_min_{ses}'] + df.loc[sub, f'NREM3_min_{ses}']
                    df.loc[sub, f'Arousal_index_NREM_{ses}'] = round(float(arou_nrem.shape[0]/(NREM_dur)*60),3)
                    
        if df.shape[1]>0: #only save dataframe if it is not empty
            df.to_csv(f'{out_dir}/{ses}_macro.csv', sep=',', index=True, index_label='ID')
        
    ### 3. Check completion status and print
    logger.info('')
    if flag == 0:
        logger.debug('Macro statistics export finished without ERROR.')  
    else:
        logger.warning('Macro statistics export finished with WARNINGS. See log for details.')
    
    return  




from setuptools import setup, find_packages
import pathlib

from os import path

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")
# Get the current version number from inside the module
with open(path.join('seapipe', 'version.py')) as version_file:
    exec(version_file.read())

setup(
    name = "seapipe",  
    version = __version__,  
    description = "Sleep Events Analysis pipeline of EEG data",  
    long_description=long_description,  
    long_description_content_type="text/markdown",  
    url="https://github.com/nathanecross/seapipe",  
    author="Nathan E. Cross",  
    author_email="nathan.cross@sydney.edu.au",
    classifiers=[  
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Research",
        "Topic :: Neuroscience :: Sleep",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords=["sleep", "electrophysiology", "detection", 
              "signal processing", "neuroscience", "analysis"], 
    packages=find_packages(),  
    python_requires=">=3.7",
    install_requires=["fooof",
                      "mne",
                      "numpy<=1.26.4",
                      "openpyxl",
                      "pandas",
                      "psutil",
                      "pyedflib",
                      "PyWavelets",
                      "pinguoin",
                      "safepickle",
                      "scipy<1.13.0",
                      "sklearn",
                      "tensorpac",
                      "wonambi",
                      "yasa"],  
    package_data={ 
        "seapipe": ["VERSION"],
    },
  
    project_urls={  # Optional
        "Bug Reports": "https://github.com/nathanecross/seapipe/issues",
        "Documentation": "https://seapipe.readthedocs.io/",
        "Source": "https://github.com/nathanecross/seapipe/",
    },
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 17 09:51:10 2025

@author: ncro8394
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

file = '/Users/ncro8394/Documents/projects/healthy/test_dataset/derivatives/bandpower_timecourse/bandpower_timeseries_0.5-4Hz_sliding/sub-HC001/ses-V1/sub-HC001_ses-V1_Fz_NREM2_sliding_0.5-4Hz.csv'

bp_df = pd.read_csv(file)



bp_ts = bp_df['band_power']

threshold = np.std(bp_ts)*10

bp_ts[bp_ts>threshold] = np.nan

plt.figure()
plt.plot(bp_ts)

n_bins = 8
bin_size = len(bp_ts) // n_bins

# Trim remainder so length divides evenly
trim = len(bp_ts) % n_bins
if trim:
    bp_ts = bp_ts[:-trim]

binned = np.nanmean(bp_ts.values.reshape(n_bins, bin_size), axis=1)



x = list(range(1,n_bins+1))
plt.bar(x, binned)




file2 = '/Users/ncro8394/Documents/projects/healthy/test_dataset/derivatives/clusterfluc/sub-HC001/ses-V0/sub-HC001_ses-V0_Fz_NREM2_SWA_fluctuations_timeseries.csv'

lf_df = pd.read_csv(file2).iloc[0]

bin_size = len(lf_df) // n_bins
# Trim remainder so length divides evenly
trim = len(lf_df) % n_bins
if trim:
    lf_df = lf_df[:-trim]

binned2 = np.nanmean(lf_df.values.reshape(n_bins, bin_size), axis=1)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 10 11:33:57 2022

@author: nathancross hello
"""

from math import floor
from numpy import where, zeros
from os import listdir, mkdir, path
from pandas import DataFrame
from wonambi import Dataset
from wonambi.attr import Annotations
from wonambi.trans import (fetch, math)


def perievent_histogram(rec_dir, xml_dir, out_dir, part, visit, stage, chan, ref_chan, cat,
                        grp_name, rater, standard, detection, chan_specific_markers,
                        window_size):

    # Make output directory
    if path.exists(out_dir):
        print(out_dir + " already exists")
    else:
        mkdir(out_dir)
    
    # Loop through channels
    for ch, channel in enumerate(chan):
        chan_ful = [channel + ' (' + grp_name + ')']
        
        # Loop through subjects and visits
        if isinstance(part, list):
            None
        elif part == 'all':
                part = listdir(xml_dir)
                part = [ p for p in part if not '.' in p]
        else:
            print("ERROR: 'part' must either be an array of subject ids or = 'all' **CHECK**")
            
        # Loop through visits
        visname = []
        if visit == 'all':
            visname = 'all'
            visit = listdir(xml_dir + '/' + part[0])
            visit = [x for x in visit if not '.' in x]
        else:
            visname.append([x for x in visit])
            
        # Make output array
        counts = zeros((len(part)*len(visit),int(window_size/0.1)))
        subvis = []
        index = -1
        
        part.sort()              
        for pp, p in enumerate(part):
            
            print(f'Running Subject {p}..')
            
            visit.sort()
            for v, vis in enumerate(visit): 
                
                # Define files
                rdir = rec_dir + '/' + p + '/' + vis + '/'
                #edf_file = [x for x in listdir(rdir) if x.endswith('.edf') if not x.startswith('.')]
                edf_file = [x for x in listdir(rdir) if x.endswith('.edf') or x.endswith('.rec') or x.endswith('.eeg')]
                xdir = xml_dir + '/' + p + '/' + vis + '/'
                xml_file = [x for x in listdir(xdir) if x.endswith('.xml') if not x.startswith('.')] 
                
                if len(xml_file) == 0:                
                    print(f'WARNING: no Annotations file found for Subject {p}, skipping..')
                else:
                    
                    # Import Annotations file
                    annot = Annotations(xdir + xml_file[0], rater_name=rater)
                    dataset = Dataset(rdir + edf_file[0])
                    s_freq = dataset.header['s_freq']
                                    
                    # Select and read data
                    print('Reading data for ' + p + ', visit ' + vis + ', channel '+ channel)
                    
                    # Retrieve standard events (to be compared against)
                    if chan_specific_markers:
                        stand_evts = fetch(dataset, annot, cat=cat, evt_type=standard, 
                                          stage=stage, cycle=None, chan_full=chan_ful, 
                                          reject_epoch=True)
                    else:
                        stand_evts = fetch(dataset, annot, cat=cat, evt_type=standard, 
                                          stage=stage, cycle=None,  reject_epoch=True)
                    
                    
                    starts = [x['times'][0][0] for x in stand_evts.segments]
                    
                    # Retrieve detection events (events to compare against standard)
                    if chan_specific_markers:
                        segments = fetch(dataset, annot, cat=cat, evt_type=detection, 
                                         stage=stage, cycle=None, chan_full=chan_ful, 
                                         reject_epoch=True)
                    else:
                        segments = fetch(dataset, annot, cat=cat, evt_type=detection, 
                                         stage=stage, cycle=None,  reject_epoch=True)
                    
                    segments.read_data([channel], ref_chan, grp_name=grp_name)
                    
                    
                    # Find the peaks of spindle events (relative to the start of the segment)
                    pktime_rel = zeros((len(segments)))
                    for i, seg in enumerate(segments):
                        dat = seg['data'] 
                        pk = math(dat, operator=_amax, axis='time').data[0][0]
                        pktime = where(seg['data'].data[0][0] == pk)[0][0]/s_freq
                        pktime_rel[i] = seg['data'].axis['time'][0][0] + pktime
                        
                    
                    histmat = zeros((len(pktime_rel),int(window_size/0.1)))
                    for y,pk in enumerate(pktime_rel):
                        for x,st in enumerate(starts):
                            
                            halfwin = window_size/2
                            # X seconds before
                            if -halfwin < pk-st < 0:
                                ind = floor((pk - st + halfwin)/0.1)
                                histmat[y,ind] = 1
                            # X seconds after
                            if 0 <= pk - st < halfwin:
                                ind = 40 + floor((pk - st)/0.1)
                                histmat[y,ind] = 1

                    index +=1
                    counts[index,:] = sum(histmat,axis=0)
                    subvis.append(p+'_'+vis)
        
        print(f"Saving histogram data to '{out_dir}peth_{channel}.csv'")
        output = DataFrame(counts, index = subvis)
        output.to_csv(f'{out_dir}peth_visit_{visname}_{channel}.csv')
                    
                        
def _amax(x, axis, keepdims=None):
    return amax(x, axis=axis)



# -*- coding: utf-8 -*-
"""
Created on Thu Jul 29 11:00:11 2021

@author: labottdv
"""

from os import listdir, mkdir, path
import shutil
from wonambi import Dataset 
from wonambi.attr import Annotations
from wonambi.detect import DetectSlowWave
from wonambi.trans import fetch

from copy import deepcopy
from datetime import datetime, date
from pandas import DataFrame
from ..utils.logs import create_logger, create_logger_outfile
from ..utils.load import (load_channels, read_inversion, load_sessions)
from ..utils.misc import infer_polarity, remove_duplicate_evts, remove_event


class seasnakes:
    
    """ Sleep Events Analysis of Slow Neocortical oscillations And K-complexES (SEASNAKES)

        This module runs a slow oscillation detection with the option of using
        previously published SO detectors:
            1. Massimini et al. (2007)
            2. Massimini et al. (2007) [Adapted to AASM criteria]
            3. Ngo et al. (2015)
            4. Staresina et al. (2015)
        
    """   
    
    def __init__(self, rec_dir, xml_dir, out_dir, chan, ref_chan, 
                 grp_name, stage, rater = None, subs='all', 
                 sessions='all', tracking = None,
                 reject_artf = ['Artefact', 'Arou', 'Arousal']):
        
        self.rec_dir = rec_dir
        self.xml_dir = xml_dir
        self.out_dir = out_dir
        self.chan = chan
        self.ref_chan = ref_chan
        self.grp_name = grp_name
        self.stage = stage
        self.rater = rater
        self.reject = reject_artf
        
        self.subs = subs
        self.sessions = sessions
        
        if tracking == None:
            tracking = {}
        self.tracking = tracking

    def detect_slowosc(self, method, cat, cycle_idx = None, duration = (0.5, 3), 
                       average_channels = False, invert = False, filetype = '.edf', 
                       logger = create_logger('Detect slow oscillations')):
        
        ''' Detect slow oscillations
        
        Runs one (or multiple) slow oscillation detection algorithms and saves the 
        detected events to a new annotations file.
        
        INPUTS:
            method ->    List of names of automated detection algorithms to detect 
                         events with. e.g. ['Massimini2004','AASM/Massimini2004',
                                            'Staresina2015', 'Ngo2015']
            cat ->       Tuple of 4 digits of either 0 or 1 e.g. (0,1,1,0) 
                         This variable sets the concatenation type when reading in 
                         the data. 0 means no concatenation, 1 means concatenation
                         #position 1: cycle concatenation
                         #position 2: stage concatenation
                         #position 3: discontinuous signal concatenation
                         #position 4: event type concatenation (does not apply here)
                         Set this based on whether you would like to detect across the
                         entire recording e.g. (1,1,1,1) or separately for each cycle
                         e.g. (0,1,1,1) or separately for each stage e.g. (1,0,1,1)
            cycle_idx->  List of indices corresponding to sleep cycle numbers.
            duration ->  Tuple of 2 digits, for the minimum and maximum duration of any 
                         detected events.     
        
        '''
        
        ### 0.a Set up logging
        flag = 0
        tracking = self.tracking
        logger.info('')
        logger.debug(r"""Commencing slow oscillation detection... 
                     
                     
                                           .•. 
                                          •   •.
                              •.         •      •       
                           .•  •        •        •.       .•.
           . .      .•.   •    •       •           •.   .•   •.¸.^8›      
              •. .•    •.•     •      •              •..        ˜¯
                               •     •                  
                               •    •                
                               •   •                  
                               •  •           
                               •.•  
                                
          
          Sleep Events Analysis of Slow Neocortical oscillations And 
                     K-complexES (S.E.A.S.N.A.K.E.S)                      
                                
                               
                                                    """,)
        ### 1. First we check the directories
        # a. Check for output folder, if doesn't exist, create
        if path.exists(self.out_dir):
                logger.debug("Output directory: " + self.out_dir + " exists")
        else:
            mkdir(self.out_dir)
        
        # b. Check input list
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = listdir(self.rec_dir)
                subs = [p for p in subs if not '.' in p]
        else:
            logger.error("'subs' must either be an array of subject ids or = 'all' ")       
        
        ### 2. Begin loop through dataset
       
        # a. Begin loop through participants
        subs.sort()
        for i, sub in enumerate(subs):
            tracking[f'{sub}'] = {}
            # b. Begin loop through sessions
            flag, sessions = load_sessions(sub, self.sessions, self.rec_dir, flag, 
                                     logger, verbose=2)
            
            for v, ses in enumerate(sessions):
                logger.info('')
                logger.debug(f'Commencing {sub}, {ses}')
                tracking[f'{sub}'][f'{ses}'] = {'slowosc':{}} 
    
                ## c. Load recording
                rdir = self.rec_dir + '/' + sub + '/' + ses + '/eeg/'
                try:
                    edf_file = [x for x in listdir(rdir) if x.endswith(filetype)]
                    dset = Dataset(rdir + edf_file[0])
                except:
                    logger.warning(f' No input {filetype} file in {rdir}')
                    break
                
                ## d. Load annotations
                xdir = self.xml_dir + '/' + sub + '/' + ses + '/'
                try:
                    xml_file = [x for x in listdir(xdir) if x.endswith('.xml')]
                    # Copy annotations file before beginning
                    if not path.exists(self.out_dir):
                        mkdir(self.out_dir)
                    if not path.exists(self.out_dir + '/' + sub):
                        mkdir(self.out_dir + '/' + sub)
                    if not path.exists(self.out_dir + '/' + sub + '/' + ses):
                        mkdir(self.out_dir + '/' + sub + '/' + ses)
                    backup = self.out_dir + '/' + sub + '/' + ses + '/'
                    backup_file = (f'{backup}{sub}_{ses}_slowosc.xml')
                    if not path.exists(backup_file):
                        shutil.copy(xdir + xml_file[0], backup_file)
                    else:
                        logger.warning(f'Annotations file already exists for {sub}, {ses}, any previously detected events will be overwritten.')
                except:
                    logger.warning(f' No input annotations file in {xdir}')
                    break
                
                # Read annotations file
                annot = Annotations(backup_file, rater_name=self.rater)
                
                ## e. Get sleep cycles (if any)
                if cycle_idx is not None:
                    all_cycles = annot.get_cycles()
                    cycle = [all_cycles[y - 1] for y in cycle_idx if y <= len(all_cycles)]
                else:
                    cycle = None
                
                ## f. Channel setup 
                pflag = deepcopy(flag)
                flag, chanset = load_channels(sub, ses, self.chan, self.ref_chan,
                                              flag, logger)
                if flag - pflag > 0:
                    logger.warning(f'Skipping {sub}, {ses}...')
                    break
                
                for c, ch in enumerate(chanset):
                    # g. Check if channel needs to be inverted for detection
                    if type(invert) == type(DataFrame()):
                        inversion = read_inversion(sub, ses, invert, ch, logger)
                        if not inversion:
                            logger.debug(f'Inferring channel polarity {ch} for {sub}, {ses}')
                            inversion = infer_polarity(dset, annot, ch, chanset[ch], 
                                                       (1,1,1,1), None, self.stage, 
                                                       cycle, logger)
                    elif type(invert) == bool:
                        inversion = invert
                    else:
                        logger.critical(f"The argument 'invert' must be set to either: 'True', 'False' or 'None'; but it was set as {invert}.")
                        logger.info('Check documentation for how to set up staging data:')
                        logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                        logger.info('-' * 10)
                        return
                    logger.debug(f"{'Inverting' if inversion else 'Not inverting'}"
                                 f" channel {ch} prior to detection for {sub}, {ses}")
                    
                    # h. Read data
                    logger.debug(f"Reading EEG data for {sub}, {ses}, {str(ch)}:{'-'.join(chanset[ch])}")
                    try:
                        segments = fetch(dset, annot, cat=cat, stage=self.stage, 
                                         cycle=cycle, reject_epoch=True, 
                                         reject_artf=self.reject)
                        segments.read_data([ch], ref_chan=chanset[ch], grp_name=self.grp_name,
                                           average_channels=average_channels)
                    except Exception as error:
                        logger.error(error.args[0])
                        logger.warning(f"Skipping {sub}, {ses}, channel {str(ch)} ... ")
                        flag+=1
                        continue
    
                    ## i. Loop through methods (i.e. Whale it!)
                    for m, meth in enumerate(method):
                        logger.debug(f'Using method: {meth}')
                        
                        # j. define detection
                        detection = DetectSlowWave(meth, duration=duration)
                        detection.invert = inversion

                        ## k. Remove any previous events on channel
                        remove_event(annot, meth, chan = None, stage = None)
                        
                        ## l. Run detection and save to Annotations file
                        if cat[0] == 1 and cat[1] == 0:
                            for s, seg in enumerate(segments):
                                logger.debug(f'Detecting events in stage {self.stage[s]}')
                                event = detection(seg['data']) # detect events
                                event.to_annot(annot, meth) # write events to annotations file
                                if len(event.events) == 0:
                                    logger.warning(f'No events detected by {meth} for {sub}, {ses}')    
                            now = datetime.now().strftime("%m-%d-%Y, %H:%M:%S")
                            tracking[f'{sub}'][f'{ses}']['slowosc'][f'{ch}'] = {'Method':meth,
                                                                              'Stage':self.stage,
                                                                              'Cycle':'All',
                                                                              'File':backup_file,
                                                                              'Updated':now}
                        else:
                            for s, seg in enumerate(segments):
                                logger.debug('Detecting events in cycle {} of {}, stages: {}'.format(s + 1, 
                                      len(segments),self.stage))
                                event = detection(seg['data']) # detect events
                                event.to_annot(annot, meth) # write events to annotations file
                                if len(event.events) == 0:
                                    logger.warning(f'No events detected by {meth} for {sub}, {ses}')
                            now = datetime.now().strftime("%m-%d-%Y, %H:%M:%S")
                            tracking[f'{sub}'][f'{ses}']['slowosc'][f'{ch}'] = {'Method':meth,
                                                                              'Stage':self.stage,
                                                                              'Cycle':list(range(1,len(segments))),
                                                                              'File':backup_file,
                                                                              'Updated':now}
                        
                        # m. Remove any duplicate detected slow oscillations on channel 
                        #remove_duplicate_evts(annot, evt_name=meth, chan=f'{ch} ({self.grp_name})')
                        
        ### 3. Check completion status and print
        if flag == 0:
            logger.info('')
            logger.debug('Slow oscillation detection finished without error.')  
        else:
            logger.info('')
            logger.warning(f'Slow oscillation finished with {flag} WARNINGS. See log for details.')
        
        #self.tracking = tracking   ## TO UPDATE - FIX TRACKING
        
        return 
    
    
    
    
    ## FOR FUTURE:
    
    # def kcomplex():
        
        
    #     return

    
class swordfish:
        
        '''
            Slow Wave Online Realtime/Rapid Detection For Intermittent Stimulation H_ (SWORDFISH)
        '''
        def __init__(self, rec_dir, xml_dir, out_dir, log_dir, chan, ref_chan, 
                 grp_name, stage, frequency=(11,16), rater = None, subs='all', 
                 sessions='all', tracking = {}):
        
            self.rec_dir = rec_dir
            self.xml_dir = xml_dir
            self.out_dir = out_dir
            self.log_dir = log_dir
            
            self.chan = chan
            self.ref_chan = ref_chan
            self.grp_name = grp_name
            self.stage = stage
            self.frequency = frequency
            self.rater = rater
            
            self.subs = subs
            self.sessions = sessions
            
            self.tracking = tracking
            
            return
    
    
    
    